% This code implements the method described in:
%
% "Consistent Partial Matching of Shape Collections via Sparse Modeling" 
% L. Cosmo, E. Rodola, A. Albarelli, F. Memoli, D. Cremers
% Computer Graphics Forum, 2016
%
% If you use this code, please cite the paper above.
% 
% Luca Cosmo, Emanuele Rodola (c) 2015

clc; clear all; close all;
addpath(genpath('./aux_functions'));
addpath(genpath('./mex'));
addpath(genpath('./geoerror'));

%% Dataset information
% params.data_dir - directory containing original models files.
% params.tmp_dir - directory containing remeshed models files
% params.classes - list of classes inside the dataset
% params.nvert - max target number of vertices of the remeshed shape

params.data_dir = './data/tosca/'; %% .off models of Tosca dataset should be put in this directory
params.tmp_dir = './data/tosca/remesh/';
params.models = {'cat', 'centaur', 'david', 'dog', 'horse', 'michael', 'victoria'};
params.model_num = {0:10, 0:5, [0 1 6 10:13], [0:3 5:8 10], [0 5:7 10 14 15 17], 0:19, [0:2 4 7 10 12 17 21 23:25]};
params.ref_shape = 1;
params.has_ref = true;
params.ext = 'off'; % original shapes model extension (vert, off)
params.nvert = 1e4;  % for remeshing

%% Setting up descriptor parameters
params.descr =         'wks'; % hks, wks
params.n_eigen =       100;
params.wks_size =      100;
params.wks_var =       6.0;

%% Setting up optimization parameters
params.n_queries = 500;
params.fps_samples = 500;
params.query_method =  'ref-fps'; %fps, fps-samp, ref-fps. See generate_query for reference.
params.samples_percentile=1;

params.sim_sigma =    3e-5;
params.metric_sigma = 10;
params.distort_fun = 'lmax';
params.validation =  'gt-rep';

%Optimization algorithm specific parameters (they should not be modified)
params.gt_min_xax =    0.5;    % termination criterion for the repeated games
params.dyn_max_iters = 5e6;
params.n_games =       100;    % max n. of L1-relaxed QAPs
params.gt_min_size =   3;      % termination criterion for gt-rep
params.dyn_error =     1e-8;
params.enet_step =     1e-4;
params.enet_alpha =    0.7;
params.min_overlap =   3;



%% Loading shapes of the specified class. 
% Shapes are remeshed according to the params.nvert value. Symmetry map and
% descriptors are computed:

cur_model = 1; %cat
shapes = load_shapes(params.models{cur_model}, params.model_num{cur_model}, params);

% Farthest points extraction  
for i=1:length(shapes)
    shapes{i}.fps = fps_euclidean(shapes{i}.VERT, params.fps_samples, 1);
end

% Run the optimization described in the paper to extract cycle-consistent 
% sparse correspondences from a collection of shapes.
[vmatches, matches,  times] = match_multiway(shapes, params);

t_queries = times.queries_matching_dyn+times.queries_matching_sim+times.query_generation;
fprintf('Total time for %d queries: %.3fs\n', params.n_queries, t_queries);
fprintf('Matches: %d, Similarity matrix computation: %.3fs, Optimization: %.3fs\n', length(vmatches), times.t_sim, times.t_opt);

% Compute the geodesic error with respect to the ground truth matches
fprintf('Computing geodesic error...');
err = calc_geo_err(vmatches, shapes, params.models, params.model_num, cur_model, params, false);
thresh = (0:0.01:0.25);
v = calc_err_curve(err, thresh);

% Plot error curve
figure('Name','Sparse correspondence')
h = plot(thresh,v);
set(h,'LineWidth', 2)
xlabel('Geodesic Error')
ylabel('% Correspondences')
set(gca,'XLim',[0 0.25])
set(gca,'YLim',[0  100])
set(gca,'PlotBoxAspectRatio',[2 1.50 1])
set(gca,'XGrid','on','YGrid','on')
title(sprintf('p0.16: %.1f\np0.02: %.1f',v(17),v(3)))
