% This file implements the method described in:
%
% "Consistent Partial Matching of Shape Collections via Sparse Modeling" 
% L. Cosmo, E. Rodola, A. Albarelli, F. Memoli, D. Cremers
% Computer Graphics Forum, 2016
%
% If you use this code, please cite the paper above.
% 
% Luca Cosmo, Emanuele Rodola (c) 2015

function descr = generate_query(shapes, params)
%% Extracts query descriptors from the shape.
%  Input:
%  - shapes: a collection of shapes
%  - params.n_queries: number of descriptors to extract
%  - params.query_method: extraction critera, possible values are:
%    * 'fps-samp': farthest point sampling on descriptor space of sampled shape points
%    * 'fps': farthest point sampling on descriptor space of all shape points
%    * 'ref-fps': farthest point sampling over the reference shape
%    * 'kmeans': K-means in descriptor space of sampled points

descr = [];
switch params.query_method    
    case 'fps-samp' % Fps on descriptors space of sampled shape points
        descrs = cell(1,numel(shapes));
        parfor i=1:numel(shapes)
            samples = shapes{i}.fps;
            switch params.descr
                case 'wks'
                    descrs{i} = shapes{i}.WKS(samples, :);
                case 'hks'
                    descrs{i} = shapes{i}.HKS(samples, :);
            end
        end
        descr = cell2mat(descrs');
        descr = descr(fps_euclidean(descr, params.n_queries, 1), :);   
    case 'ref-fps' % Farthest point sampling on a reference shape
        samples = fps_euclidean(shapes{params.ref_shape}.VERT, params.n_queries, 1);
        switch params.descr
            case 'wks'
                descr = shapes{params.ref_shape}.WKS(samples, :);
            case 'hks'
                descr = shapes{params.ref_shape}.HKS(samples, :);
        end
        
    case 'fps'     % Fps on descriptors space of all shape points
        for i=1:numel(shapes)
            switch params.descr
                case 'wks'
                    descr = [descr ; shapes{i}.WKS];
                case 'hks'
                    descr = [descr ; shapes{i}.HKS];
            end
        end
        descr = descr(fps_euclidean(descr, params.n_queries, 1), :);        
    case 'kmeans'  % K-means in descriptors space
        for i=1:numel(shapes)
            switch params.descr
                case 'wks'
                    descr = [descr ; shapes{i}.WKS(shapes{i}.fps,:)];
                case 'hks'
                    descr = [descr ; shapes{i}.HKS(shapes{i}.fps,:)];
            end
        end
        [~,descr] = fkmeans(descr, params.n_queries);   
end

assert(params.n_queries == size(descr,1))

end
