% This file implements the method described in:
%
% "Consistent Partial Matching of Shape Collections via Sparse Modeling" 
% L. Cosmo, E. Rodola, A. Albarelli, F. Memoli, D. Cremers
% Computer Graphics Forum, 2016
%
% If you use this code, please cite the paper above.
% 
% Luca Cosmo, Emanuele Rodola (c) 2015

function err = calc_geo_err(matches, shapes, models, model_num, cur_model, params, show_mismatches)
%% Computes the geodesic distance from the extracted matches with respect 
%  to the ground truth. 

% initialize fast-marching geodesics on all shapes
geo = cell(1, length(shapes));
for i=1:length(shapes)
    M = shapes{i};
    geo{i} = fastmarchmex('init', int32(M.TRIV-1), double(M.VERT(:,1)), double(M.VERT(:,2)), double(M.VERT(:,3)));
end

% load ground-truth labels
gt = cell(1, length(shapes));
sym = load(strcat(params.data_dir, 'gt/', models{cur_model}, '.sym.labels'));
for i=1:length(shapes)
    gt{i} = load(strcat(params.data_dir, 'gt/', models{cur_model}, num2str(model_num{cur_model}(i)), '.labels'));
end

err = [];

areas = zeros(1,length(shapes));
for i=1:length(shapes)
    areas(i) = sqrt(sum(shapes{i}.S_tri));
end

errors = cell(length(shapes));
errors_sym = errors;

for k=1:length(matches)
    
    m = matches{k};
    n_matched_shapes = size(m,1);
    
    for i=1:n_matched_shapes
        
        si = m(i,1);
        p = m(i,2);
        
        % this corresponds to p in the high-res reference shape
        p_ref_hires = gt{si}(shapes{si}.orig_idx(p));
        p_ref_hires_sym = sym(p_ref_hires);
        
        for j=1:n_matched_shapes
            
            if i==j 
                continue
            end
            
            sj = m(j,1);
            q = m(j,2);
            
            % note that p_ref_hires==gt{sj} gives the ground-truth match of
            % p_ref_hires
            
            % ground-truth match of p (and its symmetric)
            q_gnd = shapes{sj}.remesh_idx(p_ref_hires==gt{sj});
            q_gnd_sym = shapes{sj}.remesh_idx(p_ref_hires_sym==gt{sj});
            
            source = inf(shapes{sj}.n, 1);
            source(q) = 0;
            d = fastmarchmex('march', geo{sj}, double(source));
            
            errors{si,sj} = [errors{si,sj} , d(q_gnd)/areas(sj)];
            errors_sym{si,sj} = [errors_sym{si,sj} , d(q_gnd_sym)/areas(sj)];
            
%             if show_mismatches
%                 % This check is motivated by possible remeshing errors that
%                 % could induce large errors even in case of exact match.
%                 if geo_err(i,j) >= 0.4
%                     whites = [0.8 0.8 0.8];
%                     figure, colormap(whites)
%                     subplot(1,2,1)
%                     plot_scalar_map(shapes{si}, ones(1,shapes{si}.n)), axis off; light; lighting phong; camlight head;
%                     hold on, plot_cloud(shapes{si}, p)
%                     subplot(1,2,2)
%                     plot_scalar_map(shapes{sj}, ones(1,shapes{sj}.n)), axis off; light; lighting phong; camlight head;
%                     hold on, plot_cloud(shapes{sj}, q), plot_cloud(shapes{sj}, [q_gnd q_gnd_sym], 'b*')
%                     title(sprintf('Error(%d,%d): %.3f', i, j, geo_err(i,j)))
%                 end
%             end
            
        end
    end
    
end % next multi-match

% For each pair of shapes, keep the better results out of the true and
% symmetric ground truths. This choice is NOT done at the level of a single
% match, but at the level of the whole correspondence between the two
% shapes.

for i=1:length(shapes)-1
    for j=i+1:length(shapes)
        e = (errors{i,j} + errors{j,i})./2; % average (p,q) and (q,p)
        esym  = (errors_sym{i,j} + errors_sym{j,i})./2;
        if sum(e) < sum(esym)
            err = [err e];
        else
            err = [err esym];
        end
    end
end

% delete fast-marching geodesics
for i=1:length(geo)
    fastmarchmex('deinit', geo{i});
end
clear geo

end
