% This file implements the method described in:
%
% "Consistent Partial Matching of Shape Collections via Sparse Modeling" 
% L. Cosmo, E. Rodola, A. Albarelli, F. Memoli, D. Cremers
% Computer Graphics Forum, 2016
%
% If you use this code, please cite the paper above.
% 
% Luca Cosmo, Emanuele Rodola (c) 2015

function shapes = load_shapes(model, model_nums, params)
%% LOAD_SHAPES loads the shapes of the specified class.
%  Input:
%   - params.data_dir: directory containing original models files. (.vert/.tri
%                   or .off) 
%   - params.tmp_dir: directory contining remeshed models files
%   - class: class name of the shapes to be loaded
%   - model_nums: shape index of the shapes to be loaded
%  Output:
%   - shapes - a cell of structures contining the following shapes inforation:
%     * shapes{i}.VERT: vertices coordinates
%     * shapes{i}.TRIV: triangles indices
%     * shapes{i}.orig_idx: map from remeshed idxes to full resolution idxes
%     * shapes{i}.remesh_idx: map from full resolution idxes to remeshed idxes 
%     * shapes{i}.WKS: Wave Kernel Signatures
%     * shapes{i}.HKS: Heat Kernel Signatures
%     * shapes{i}.sym: Left-Right labelling

if ~exist(params.tmp_dir, 'dir'); mkdir(params.tmp_dir); end; 

shapes = cell(1, length(model_nums));

% computing the simmetry map on reference shape
if params.has_ref
    prefix = strcat(params.data_dir, model, '0');
    Msym = {};
    switch params.ext
        case 'vert'
            Msym = {};
            Msym.VERT = load(strcat(prefix,'.vert'));
            Msym.TRIV = load(strcat(prefix,'.tri'));
        case 'off'
            Msym = load_off(strcat(prefix,'.off'));
        otherwise
            error('Unrecognized shape format.');
    end

    pleft = Msym.VERT(:,1) < 0;
    pright = ~pleft;
    sym = double(pleft) - double(pright);
end


for i=1:length(model_nums)
    
    fprintf('Loading shape %d/%d...\n', i, length(model_nums));
    prefix = strcat(params.data_dir, model, num2str(model_nums(i)));
    
    out_fname = strcat(params.tmp_dir, model, num2str(model_nums(i)), '.mat');
    
    if ~exist(out_fname, 'file')
        
        
        switch params.ext
            case 'vert'
                M = {};
                M.VERT = load(strcat(prefix,'.vert'));
                M.TRIV = load(strcat(prefix,'.tri'));
            case 'off'
                M = load_off(strcat(prefix,'.off'));
            otherwise
                error('Unrecognized shape format.');
        end

        M_remeshed = remesh(M, struct('verbose',false,'vertices',params.nvert));
        
        % keep a LUT to the original vertex indices
        tree = ann('init', M.VERT');
        M_remeshed.orig_idx = ann('search', tree, M_remeshed.VERT', 1);
        ann('deinit', tree); clear tree;
        
        % map from high-res to remeshed shape
        tree = ann('init', M_remeshed.VERT');
        M_remeshed.remesh_idx = ann('search', tree, M.VERT', 1);
        ann('deinit', tree); clear tree;
        
        M = M_remeshed;
        M.n = size(M.VERT,1);
        M.m = size(M.TRIV,1);
        
        if params.has_ref
            % symmetry information
            M.sym = sym(M.orig_idx);
        end

        % LB basis + area elements
        [M.evecs, M.evals, M.S] = calc_LB(M, params.n_eigen);

        % descriptors
        M.WKS = calc_WKS(M, params.wks_size, params.wks_var);
        M.HKS = calc_HKS(M);
        
        % exact triangle areas
        M.S_tri = calc_tri_areas(M);
        
        save(out_fname, 'M');        
    else
        M = load(out_fname);
        M = M.M;
    end
    
    shapes{i} = M;
end

end
