% This file implements the method described in:
%
% "Consistent Partial Matching of Shape Collections via Sparse Modeling" 
% L. Cosmo, E. Rodola, A. Albarelli, F. Memoli, D. Cremers
% Computer Graphics Forum, 2016
%
% If you use this code, please cite the paper above.
% 
% Luca Cosmo, Emanuele Rodola (c) 2015


function [vmatches, matches, times] = match_multiway(shapes, params)
%% MATCH_MULTIWAY implements the optimization method described in 
%  <"Consistent Partial Matching of Shape Collections via Sparse Modeling",
%  L. Cosmo, E. Rodola, A. Albarelli, F. Memoli, D. Cremers, CGF 2016> to
%  extract cycle-consistent sparse correspondences from a collection of
%  possibly heterogeneous shapes.  
%  Input.
%  - shapes: a collection of shapes
%  - params: optimization parameters
%  Output:
%  - vmatches: the final optimized multiway-matches
%  - matches: the matches belonging to the reduced feasible set (after first
%            sub-problem optimization)
%  - times: optimization times
    
    %%
    % -------------------------------------------------------------------------
    % First sub-problem (section 4.1).
    % With this step the optimization space of the second sub-problem is 
    % reduced by extracting only few consistent multiway-matches maximizing
    % some measure of point similarity. 
    % In order to promote a correspondence better distributed over the whole
    % shape surface, some evenly distributed (over the surface or the 
    % descriptor space) seed queries are selected and used to extract
    % single multiway matches from a subset of shape points similar to 
    % the query.
    % -------------------------------------------------------------------------
    
    tic;
    %%Extracts query descriptors from the shape.
    queries = generate_query(shapes, params);
    params.n_queries = size(queries,1);
    times.query_generation = toc;
    
    matches = cell(1, params.n_queries);
    fprintf('Matching %d queries...\n',size(queries,1));
    
    sum_t_sim = 0;
    sum_t_dyn = 0;
    tic;
    parfor q=1:size(queries,1)
        %The first optimization sub-problem is repeated for each query
        %to extract cycle-consistent matches
        [m, t_sim, t_dyn] = match_query(queries(q,:), shapes, params);
        sum_t_sim = sum_t_sim+t_sim;
        sum_t_dyn = sum_t_dyn+t_dyn;
        matches{q} = m;    
    end

    discard = false(params.n_queries,1);
    for q=1:size(queries,1)
       if(size(matches{q},1)==0); discard(q)=1; end; 
    end
    matches = matches(~discard);
    params.n_queries = numel(matches);
        
    elapsed = toc;
    rapp = sum_t_sim/(sum_t_sim+sum_t_dyn);
    times.queries_matching_sim = elapsed*rapp;
    times.queries_matching_dyn = elapsed*(1-rapp);
    
    % -------------------------------------------------------------------------
    % Second sub-problem (section 4.2).
    % This step keeps only multiway-matches minimizing the metric distortion
    % -------------------------------------------------------------------------
   
    [vmatches, times.t_sim, times.t_opt] = validate_matches(matches, shapes, params);
end
