% This file implements the method described in:
%
% "Consistent Partial Matching of Shape Collections via Sparse Modeling" 
% L. Cosmo, E. Rodola, A. Albarelli, F. Memoli, D. Cremers
% Computer Graphics Forum, 2016
%
% If you use this code, please cite the paper above.
% 
% Luca Cosmo, Emanuele Rodola (c) 2015

function [mw_match, t_sim, t_dyn] = match_query(query, shapes, params)
%% MATCH_QUERY solves for a single multiway match of the first sub-problem
%  as described in section 4.1 of the implemented paper.      
%  Input:
%  - query: a point in the descriptor space
%  - shapes: a collection of shapes
%  - params: structur containing algorithm and optimization parameters.
%  Output:
%  - mw_match: a cell containing the optimized multiway-matches. Each mw-match
%              is a matrix contining for each row a shape index and the 
%              relative point index of the mw-match.  

    tic;
    %extract from each shape points with a descriptor similar to the query
    samples = get_shape_samples(shapes, query, params);


    samples_descr = cell(1, length(shapes));
    for i=1:length(shapes)
        switch params.descr
            case 'wks'
                samples_descr{i} = shapes{i}.WKS(samples{i},:)';
            case 'hks'
                samples_descr{i} = shapes{i}.HKS(samples{i},:)';
            otherwise
                error('Unknown descriptor.');
        end
    end

    % Construct the distance matrix
    D = build_similarity_matrix(samples_descr);

    % When the left-right maps of the shapes are given, they can be used to
    % obtain orientation-preserving solutions by setting zero similarity to
    % inconsistent candidate matches.
    sym_mask = false(size(D));
    for i=1:length(shapes)-1
        for j=i+1:length(shapes)
            mask = repmat(shapes{i}.sym(samples{i}), 1, length(samples{j})) == repmat(shapes{j}.sym(samples{j})', length(samples{i}), 1);
            from_row = (i-1)*length(samples{i})+1;
            to_row = from_row+length(samples{i})-1;        
            from_col = (j-1)*length(samples{j})+1;
            to_col = from_col+length(samples{j})-1;
            sym_mask(from_row:to_row, from_col:to_col) = mask;
        end
    end
    sym_mask = sym_mask | sym_mask';




    %%Transorm the distance matrix to the similarity matrix with a Gaussian
    % weight (eq 14);
    sigma2 = params.sim_sigma^2;
    S = exp(-(1/sigma2)*D);
    % set intra shape matches similarity to zero to avoid selecting multiple points from the same shape
    z = (D==0); % these should *only* be the diagonal 0-blocks
    S(z) = 0;

    S(~sym_mask) = 0; %apply left-right mask if any given

    t_sim = toc;

    assert(~any(any(isnan(S))),'NaN in the similarity matrix');
    assert(~any(any(isinf(S))),'Inf in the similarity matrix');

    % Solve the L1-relaxed QAP
    tic;
    x = ones(size(S,1),1) + rand(size(S,1),1)*0.1;
    x = x' / sum(x);

    [dyn_iters, dyn_err, dyn_x] = iidyn(S, x, params.dyn_error, params.dyn_max_iters);
    t_dyn = toc;

    if dyn_iters==0
        error('The dynamics did not even start.')
    end

    if dyn_iters==params.dyn_max_iters
        error('The dynamics did not reach convergence (max iterations reached).')
    end

    if sum(dyn_x>0) > length(shapes)
        imagesc( exp(-(1/1e-7)*sp));
        warning('One-to-one constraints are not satisfied.') 
        mw_match = [];
        return;
    end

    % Convert back to multi-way correspondence, i.e. list of (shape,point_id)
    idx_to_sample = [];
    for i=1:length(shapes)
        idx_to_sample = [idx_to_sample ; i*ones(length(samples{i}),1) , reshape(samples{i}, numel(samples{i}),1) ];
    end

    mw_match = idx_to_sample(dyn_x>0, :);

end
