/*
 * This file implements the method described in:
 *
 * "Consistent Partial Matching of Shape Collections via Sparse Modeling"
 * L. Cosmo, E. Rodola, A. Albarelli, F. Memoli, D. Cremers
 * Computer Graphics Forum, 2016
 *
 * If you use this code, please cite the paper above.
 * 
 * Luca Cosmo, Emanuele Rodola (c) 2015
 *
 * To compile:
 * mex build_similarity_matrix.cpp COMPFLAGS="/openmp $COMPFLAGS"
 */

#include "mex.h"
#include <vector>

//DEBUG
#include <fstream>
#include <iostream>

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[])
{
	if (nrhs != 1 || nlhs != 1)
      mexErrMsgTxt("Usage: D = build_similarity_matrix(descriptors).");

	const double* const descr = mxGetPr(prhs[0]);
	const int n_shapes = std::max(int( mxGetM(prhs[0]) ) , int( mxGetN(prhs[0]) ));
	
	int tot_samples = 0;
	for (int i=0; i<n_shapes; ++i)
	{
		mxArray* shape = mxGetFieldByNumber(prhs[0], 0, i);
		if (shape==0)
			mexErrMsgTxt("Error when parsing the descriptors struct.");
		else
		{
			const int ns = mxGetN(shape);
			tot_samples += ns;
		}
	}
	
	plhs[0] = mxCreateDoubleMatrix(tot_samples, tot_samples, mxREAL);
	double* D = mxGetPr(plhs[0]);
	
	std::fill(D, D+tot_samples*tot_samples, 0.);
	
	int row = 0, col = 0; // in the output matrix
	
	for (int s1=0; s1<n_shapes-1; ++s1)
	{
		mxArray* shape1 = mxGetFieldByNumber(prhs[0], 0, s1);
		const double* descr1 = mxGetPr(shape1);
		const int n1 = mxGetN(shape1);
		
		const int sz = mxGetM(shape1);

		for (int i=0; i<n1; ++i)
		{
			const double* d1 = descr1 + i*sz;
			
			col = 0;
			for (int ii=0; ii<s1; ++ii)
				col += mxGetN( mxGetFieldByNumber(prhs[0], 0, ii) );
					
			for (int s2=s1+1; s2<n_shapes; ++s2)
			{
				mxArray* shape2 = mxGetFieldByNumber(prhs[0], 0, s2);
				const double* descr2 = mxGetPr(shape2);
				const int n2 = mxGetN(shape2);
				
				col += n2;
				
				for (int j=0; j<n2; ++j)
				{
					const double* d2 = descr2 + j*sz;
					
					double dist = 0.;
					
					for (int k=0; k<sz; ++k)
					{
						const double diff = d1[k] - d2[k];
						dist += diff*diff;
					}
					
					D[row*tot_samples+(col+j)] = D[(col+j)*tot_samples+row] = dist;
					//std::cout << row << " , " << col+j << std::endl;
				
				} // next sample in the second shape
				
			} // next shape
			
			++row;

		} // next sample in the first shape
		
	} // next shape
}
