/*
 * This file implements the method described in:
 *
 * "Consistent Partial Matching of Shape Collections via Sparse Modeling"
 * L. Cosmo, E. Rodola, A. Albarelli, F. Memoli, D. Cremers
 * Computer Graphics Forum, 2016
 *
 * If you use this code, please cite the paper above.
 * 
 * Luca Cosmo, Emanuele Rodola (c) 2015
 *
 * To compile:
 * mex prune_mw_matches.cpp
 */

#include "mex.h"
#include <vector>

//DEBUG
#include <fstream>
#include <iostream>
#include <algorithm>

typedef std::vector< std::pair<int,int> > multiway_match; // list of <shape_id,point_id>

void prune_matches(const double* x, const std::vector<multiway_match>& candidate_matches, std::vector<int>& matches)
{
	const int n = (int)candidate_matches.size();

	std::vector<bool> is_active(n, true);
	
	int best_match = -1;
	double best_match_val = 0.;

	for (int i=0; i<n; ++i)
	{
		const double xi = x[i];
		if (xi == 0.)
			is_active[i] = false;
		if (xi > best_match_val)
		{
			best_match_val = xi;
			best_match = i;
		}
	}

	while (best_match != -1)
	{
		is_active[best_match] = false;
		matches.push_back(best_match);
		
		const multiway_match& leader = candidate_matches[best_match];
		
		for (int i=0; i<n; ++i)
		{
			if (is_active[i])
			{
				const multiway_match& m = candidate_matches[i];
				bool is_conflicting = false;
				
				for (size_t j=0; j<m.size(); ++j)
				{
					for (size_t k=0; k<leader.size(); ++k)
					{
						if ( m[j] == leader[k] )
						{
							is_conflicting = true;
							//std::cout << i << "," << best_match << std::endl;
							//std::cout << "(" << m[j].first << "," << m[j].second << ") and (" << leader[k].first << "," << leader[k].second << ")" << std::endl;
							break;
						}
					}
					if (is_conflicting)
                    {
                        //mexPrintf("Conflicting!\n");
						break;
                    }
				}
				
				is_active[i] = !is_conflicting;
			}
		}

		best_match = -1;
		best_match_val = 0.;
		
		for (int i=0; i<n; ++i)
		{
			if (is_active[i])
			{
				const double xi = x[i];
				if (xi > best_match_val)
				{
					best_match_val = xi;
					best_match = i;
				}
			}
		}
	}
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[])
{
	if (nrhs != 2 || nlhs != 1)
      mexErrMsgTxt("Usage: matches_idx = prune_mw_matches(x, candidate_matches).");

	const double* const x = mxGetPr(prhs[0]);
	const int n = std::max(int( mxGetM(prhs[0]) ) , int( mxGetN(prhs[0]) ));
	
	std::vector<multiway_match> candidate_matches(n);
	
	for (int i=0; i<n; ++i)
	{
		mxArray* fld = mxGetFieldByNumber(prhs[1], 0, i);
		if (fld==0)
			mexErrMsgTxt("Error when parsing the matches struct.");
		else
		{
			mxClassID type = mxGetClassID(fld);
			if (type != mxDOUBLE_CLASS)
				mexErrMsgTxt("The candidate_matches cell must contain matrices of type double.");
			
			double* p = mxGetPr(fld);
			
			if (mxGetN(fld) != 2)
				mexErrMsgTxt("The matches structure must be of cell type, containing two-column matrices.");
			
			const int nm = mxGetM(fld);
			
			multiway_match& m = candidate_matches[i];
			m.resize(nm);
			
			double* ptr = mxGetPr(fld);
			for (int j=0; j<nm; ++j)
			{
				m[j].first = ptr[j];     // shape id
				m[j].second = ptr[j+nm]; // point id (both are 1-based)
			}
		}
	}
			
	std::vector<int> matches;
	prune_matches(x, candidate_matches, matches);
	
	plhs[0] = mxCreateDoubleMatrix(matches.size(), 1, mxREAL);
    double* p = mxGetPr(plhs[0]);
	
	for (size_t i=0; i<matches.size(); ++i)
		p[i] = static_cast<double>(matches[i])+1;
}
