/*
 * This file implements the method described in:
 *
 * "Consistent Partial Matching of Shape Collections via Sparse Modeling"
 * L. Cosmo, E. Rodola, A. Albarelli, F. Memoli, D. Cremers
 * Computer Graphics Forum, 2016
 *
 * If you use this code, please cite the paper above.
 * 
 * Luca Cosmo, Emanuele Rodola (c) 2015
 *
 * To compile:
 * mex track_similarity.cpp COMPFLAGS="/openmp $COMPFLAGS"
 */
 
#include <limits>
#include <iostream>
#include "mex.h"
#include <algorithm>

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[])
{ 
    if (nrhs != 5)
		mexErrMsgTxt("Usage:: [S] = track_similarity(M,D,min_overlap, sigma, method)");
    if (nlhs > 1)
		mexErrMsgTxt("Too many output arguments."); 

	// input
	const int *M = (int*)mxGetPr(prhs[0]);	
    const mxArray **Dcell = (const mxArray **)mxGetPr(prhs[1]);	
    
    const int n_match = int( mxGetN(prhs[0]));
    const int n_shapes = int( mxGetN(prhs[1]));
    
    const int overlap = int(*mxGetPr(prhs[2]));
    const double sigma = double(*mxGetPr(prhs[3]));
    double sigma2 =sigma*sigma;
  
    char command[1024];
	int strlen = (int)mxGetN(prhs[4]);
	strlen = (int)(strlen < mxGetM(prhs[4]) ? mxGetN(prhs[4]) : strlen);
	strlen += 1;
	mxGetString (prhs[4], command, strlen);
        
    //Decode Method
    int method = -1;    
    if (!strcmp(command, "lmax")) { method = 0; }
    if (!strcmp(command, "l2")) { method = 1; }
    if (!strcmp(command, "l3")) { method = 2; }
    if (!strcmp(command, "relative")) { method = 3; }
    if (!strcmp(command, "consistency")) { method = 4; }
    
    //Extract distance matrices for each shape
    int *DsSize = new int[n_shapes];
    double **Ds = new double*[n_shapes];
    for(int i=0;i<n_shapes;i++)
        Ds[i]=mxGetPr(Dcell[i]);
    for(int i=0;i<n_shapes;i++)
        DsSize[i]=int( mxGetN(Dcell[i]));
    
    // output
	plhs[0] = mxCreateDoubleMatrix( (mwSize)n_match, (mwSize)n_match, mxREAL);
	double* S = mxGetPr(plhs[0]);
	   
    
	// run	

    #pragma omp parallel for
    for(int i=0; i<n_match; ++i)
        for(int j=i+1;j<n_match;++j)
        {   
            S[i*n_match+j] = S[j*n_match+i] = 0;
            
            //overlap
            int overlapping=0;
            double min_dist=999999999999999;            
            double max_dist=0;
            
            double *dists = new double[n_shapes];
            for(int m=0; m<n_shapes; m++)
            {
                dists[m] = -1;
                if(M[i*n_shapes + m]>0 && M[j*n_shapes + m]>0)
                {
                   if(M[i*n_shapes + m]==M[j*n_shapes + m])
                   {
                       S[i*n_match+j] = S[j*n_match+i] = 0;
                       overlapping=-1;
                       break;
                   }
                 
                   double dist = (Ds[m][(M[i*n_shapes + m]-1)*DsSize[m]+(M[j*n_shapes + m]-1)]);
                   dists[m] = dist;
                   
                   min_dist=std::min(dist,min_dist);
                   max_dist=std::max(dist,max_dist);
                   overlapping++;
                }
            }
            if(overlapping>=overlap) 
            {
                double max_diff=0;
                double sum_diff2=0;
                double sum_diff3=0;
            
                for(int di = 0; di < n_shapes; ++di)
                    for(int dj = di+1; dj < n_shapes; ++dj)
                    {
                        if(dists[di] == -1 || dists[dj] == -1) continue;
                        double diff = std::abs(dists[di]-dists[dj]);

                        max_diff = std::max(diff,max_diff);
                        sum_diff2+=diff*diff;
                        sum_diff3+=diff*diff*diff;
                    }
                
                switch(method) {
                    case (0):
                       S[i*n_match+j]= S[j*n_match+i] = std::exp( (-1.0/sigma2)*max_diff*max_diff);
                       break;
                    case (1):
                       S[i*n_match+j]= S[j*n_match+i] = std::exp( (-1/sigma2)*sum_diff2);
                       break;
                    case (2):
                       S[i*n_match+j]= S[j*n_match+i] = std::exp( (-1/sigma2)*std::pow(sum_diff3,2.0/3));
                       break;
                    case (3):
                       S[i*n_match+j]= S[j*n_match+i] = std::pow((min_dist/max_dist),sigma);
                       break;
                    case (4):
                       S[i*n_match+j]= S[j*n_match+i] = 1;
                    break;
                    default:
                        mexErrMsgTxt("Wrong method."); 
                        
                }
            }
            if(method == 4 && overlapping>=0)
                S[i*n_match+j]= S[j*n_match+i] = 1;
            
            delete[] dists;
        }
    
    delete[] Ds;
    delete[] DsSize;
    return;
}
