% This file implements the method described in:
%
% "Consistent Partial Matching of Shape Collections via Sparse Modeling" 
% L. Cosmo, E. Rodola, A. Albarelli, F. Memoli, D. Cremers
% Computer Graphics Forum, 2016
%
% If you use this code, please cite the paper above.
% 
% Luca Cosmo, Emanuele Rodola (c) 2015

function [vmatches, t_sim, t_opt] = validate_matches(matches, shapes, params)
%% VALIDATE_MATCHES solves extracts from the set of mw-matches given as 
%  input the subset of mw-matches minimizing the metric distortion as 
%  described in section 4.2 of the implemented paper.      
%  Input:  
%  - matches: a cell array containing mw-matches
%  - shapes: a collection of shapes
%  - params: optimization parameters
%  Output:
%  - mw_match: a cell containing the optimized multiway-matches. Each mw-match
%              is a matrix containing for each row a shape index and the 
%              relative point index of the mw-match.  

fprintf('Validating multi-way matches...\n ');
tic;
T = build_validation_matrix(shapes, matches, params);
t_sim = toc;
%T = T+1e-3*mean(T(:))*rand(size(T));

tic
t_opt = 0;
switch params.validation
    
    case 'gt' % Game-theoretic (L1) solution
        
        x = ones(length(matches),1);
        x = x' / sum(x);
        
        tic;
        [out_iters, out_err, out_x] = iidyn(T, x, params.dyn_error, params.dyn_max_iters);
        t_opt = toc;
        
        if out_iters==params.dyn_max_iters
            error('The dynamics did not reach convergence (max iterations reached).')
        end
        
        vmatches = matches(out_x>0);
        
    case 'gt-rep' % Repeated game-theoretic

        tic;
        groups = cell(1,params.n_games);
        tot_iters = 0;
        avg_err = 0;
        
        for iter=1:params.n_games
%           iter
            x = ones(length(matches),1)+rand(length(matches),1)*0.01;
            x = x' / sum(x);

            [out_iters, out_err, out_x] = iidyn(T, x, params.dyn_error, params.dyn_max_iters);
            
            obj = out_x*T*out_x';
            sol = find(out_x);
            
            if iter > 1 % this check ensures that we get a solution in any case
                if obj < params.gt_min_xax || length(sol) <= params.gt_min_size
                    iter = iter - 1;
                    break
                end
            end
            
            tot_iters = tot_iters + out_iters;
            avg_err = avg_err + out_err;
            
            groups{iter} = sol;
            
            T(sol,:) = 0;
            T(:,sol) = 0;
                       
        end % next game

        fprintf('%d L1-relaxed QAPs\n', iter);

        grouped_matches = matches(cat(2,groups{:}));
        t_opt = toc;
        out_iters = tot_iters;
        out_err = avg_err / iter;
        
        tic
        S = build_validation_matrix(shapes, grouped_matches, params);
        t_sim = t_sim+toc;
        
        tic
        [sol_l2,ev] = eigs(S,1);
        sol_l2 = abs(sol_l2);
        matches_idx = prune_mw_matches(sol_l2, grouped_matches);
        t_opt = t_opt + toc;
        vmatches = grouped_matches(matches_idx);
         
        [sol_l2,ev] = eigs(S,1);
        grouped_matches = grouped_matches(abs(sol_l2)>10e-10);
    case 'spectral' % Spectral (L2) solution
        
        tic;
        
        [sol_l2,~] = eigs(T,1);
        sol_l2 = abs(sol_l2);
        matches_idx = prune_mw_matches(sol_l2, matches);
        
        t_opt = toc;
        
        vmatches = matches(matches_idx);
        out_iters = 0;
        out_err = 0;
        mask = false(length(matches),1);
        mask(matches_idx) = true;
        out_x = sol_l2;
        out_x(~mask) = 0;
        
    case 'enet' % Elastic net (L1+L2) solution
        
        tic;
        
        [sol_enet, out_iters, out_err] = enet_grad_desc(T, params.enet_alpha, [], params.dyn_error, params.dyn_max_iters, params.enet_step);
        sol_enet(sol_enet<0.05*max(sol_enet)) = 0;
        matches_idx = prune_mw_matches(sol_enet, matches);
        
        t_opt = toc;
        
        vmatches = matches(matches_idx);
        mask = false(length(matches),1);
        mask(matches_idx) = true;
        out_x = sol_enet;
        out_x(~mask) = 0;
        
    otherwise
        error('Invalid validation method.');
end

fprintf('done.\n');

end
